/**********************************************************************
Copyright (c) 2008 Andy Jefferson and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contributors:
    ...
**********************************************************************/
package org.datanucleus.store.excel;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.Map;

import javax.transaction.xa.XAResource;

import org.apache.poi.ss.usermodel.Workbook;
import org.datanucleus.ExecutionContext;
import org.datanucleus.exceptions.NucleusException;
import org.datanucleus.store.StoreManager;
import org.datanucleus.store.connection.AbstractConnectionFactory;
import org.datanucleus.store.connection.AbstractManagedConnection;
import org.datanucleus.store.connection.ManagedConnection;

/**
 * Implementation of a ConnectionFactory for Excel.
 */
public class ConnectionFactoryImpl extends AbstractConnectionFactory
{
    String filename = null;
    boolean ooxml = false;

    /**
     * Constructor.
     * @param storeMgr Store Manager
     * @param resourceType Type of resource (tx, nontx)
     */
    public ConnectionFactoryImpl(StoreManager storeMgr, String resourceType)
    {
        super(storeMgr, resourceType);
        // "excel:file:{filename}"/"ooxml:file:{filename}"
        String url = storeMgr.getConnectionURL();
        if (url == null)
        {
            throw new NucleusException("You haven't specified persistence property 'datanucleus.ConnectionURL' (or alias)");
        }

        if (url.startsWith("excel:"))
        {
            ooxml = false;
        }
        else if (url.startsWith("ooxml:"))
        {
            ooxml = true;
        }
        else
        {
            throw new NucleusException("invalid URL: "+url);
        }

        // Split the URL into filename
        String str = url.substring(6); // Omit "excel:"/"ooxml:"
        if (str.indexOf("file:") != 0)
        {
            throw new NucleusException("invalid URL: "+url);
        }

        filename = str.substring("file:".length()); // Omit "file:"
    }

    /**
     * Obtain a connection from the Factory. The connection will be enlisted within the transaction
     * associated to the ExecutionContext
     * @param ec the pool that is bound the connection during its lifecycle (or null)
     * @param txnOptionsIgnored Any options for creating the connection
     * @return the {@link org.datanucleus.store.connection.ManagedConnection}
     */
    public ManagedConnection createManagedConnection(ExecutionContext ec, Map txnOptionsIgnored)
    {
        return new ManagedConnectionImpl();
    }

    public class ManagedConnectionImpl extends AbstractManagedConnection
    {
        /** The Excel XLS/OOXML file. */
        File file;
        
        public ManagedConnectionImpl()
        {
        }

        public Object getConnection()
        {
            if (conn == null)
            {
                try
                {
                    file = new File(filename);
                    if (!file.exists())
                    {
                        // Excel document doesn't exist, so create
                        Workbook wb = null;
                        if (ooxml)
                        {
                            wb = new org.apache.poi.xssf.usermodel.XSSFWorkbook();
                        }
                        else
                        {
                            wb = new org.apache.poi.hssf.usermodel.HSSFWorkbook();
                        }
                        FileOutputStream fileOut = new FileOutputStream(file);
                        wb.write(fileOut);
                        fileOut.close();
                    }

                    InputStream is = new FileInputStream(file);
                    if (ooxml)
                    {
                        conn = new org.apache.poi.xssf.usermodel.XSSFWorkbook(is);
                    }
                    else
                    {
                        conn = new org.apache.poi.hssf.usermodel.HSSFWorkbook(is);
                    }
                }
                catch (IOException e)
                {
                    throw new NucleusException(e.getMessage(),e);
                }
            }
            return conn;
        }

        public void release()
        {
            if (commitOnRelease)
            {
                // Non-transactional operation end : Write to file and close connection
                try
                {
                    FileOutputStream os = new FileOutputStream(file);
                    ((Workbook)conn).write(new FileOutputStream(file));
                    os.close();
                    file = null;
                    conn = null;
                }
                catch (IOException e)
                {
                    throw new NucleusException(e.getMessage(),e);
                }
            }
            super.release();
        }

        public void close()
        {
            if (conn == null)
            {
                return;
            }

            try
            {
                for (int i=0; i<listeners.size(); i++)
                {
                    listeners.get(i).managedConnectionPreClose();
                }

                FileOutputStream os = new FileOutputStream(file);
                ((Workbook)conn).write(new FileOutputStream(file));
                os.close();
                file = null;
                conn = null;
            }
            catch (IOException e)
            {
                throw new NucleusException(e.getMessage(),e);
            }
            finally
            {
                for (int i=0; i<listeners.size(); i++)
                {
                    listeners.get(i).managedConnectionPostClose();
                }
            }
        }

        public XAResource getXAResource()
        {
            return null;
        }
    }
}